/**
 * A simple compiler to translate e-sql(the sql contains variable placeholders)
 * to js-procedure.
 * @Author zyz.
 */

// SQL result regex.
var rReg = /\$\d+((\.\w+)|(\[\s*\d+\s*\])|(\[\s*'[^']+'\s*\])|(\[\s*"[^"]+"\s*\]))+/;

// SQL identifier regex.
var iReg = /\$(\d+-)?\w+(\.\w+)*\$/;

// SQL parameter regex.
var pReg = /\@(\d+-)?\w+(\.\w+)*\@/;

// All types of placeholder regex
var reg = /(\$(\d+-)?\w+(\.\w+)*\$)|(\@(\d+-)?\w+(\.\w+)*\@)|(\$\d+((\.\w+)|(\[\s*\d+\s*\])|(\[\s*'[^']+'\s*\])|(\[\s*"[^"]+"\s*\]))+)/g;

var expReg = /\#([^\#]|\s)*\#/g;

// Identifiers and parameters
var ipReg = /(\@(\d+-)?\w+(\.\w+)*\@)|(\$(\d+-)?\w+(\.\w+)*\$)/g;

var procedureTemplateTransaction = require('./procedure-template-transaction.js');
var procedureTemplate = require('./procedure-template.js');

var compiler = {
    compile: function(eSql) {
        // Remove blank.
        eSql = eSql.replace(/(^\s+)|(\s+$)/g, '');
        // Remove comments.
        eSql = eSql.replace(/(\/\*(.|\s)+?\*\/)|(--.*)/g, '');

        eSql = eSql.replace(expReg, function(s) {
            return s.replace(/\@/g, '{^}')
                .replace(/\$/g, '{&}')
                .replace(/\;/g, '{%}');
        });
        if (!eSql) {
            return;
        }
        var series = false;
        var sideEffectsSQLCount = 0;
        var esqlSet = eSql.split(/;/g)
        .map(function(sql) {
            return sql.replace(/^\s|\s$/g, '');
        })
        .filter(function(sql) {
            if (sql && !/^SELECT\s/i.test(sql)) {
                sideEffectsSQLCount++;
            }
            return sql ? true : false;
        });

        var fragments = esqlSet.map(function(sql, index) {
            if (!sql) {
                return '';
            }
            var fragment = [];
            var params = [];
            var dependencies = [];
  
            sql = sql.replace(reg, function(s) {
                return '#' + s + '#';
            })
            .replace(/\{\^\}/g, '@')
            .replace(/\{\&\}/g, '$')
            .replace(/\{\%\}/g, ';');

            sql = '"' + sql + '"';
            sql = sql
            .replace(expReg, function(s) {
                if (rReg.test(s)) {
                    series = true;
                }

                var exp = s.replace(/\#/g, '').replace(/(^\s*)|(\s*$)/g, '');
                var p = '(function(){return '
                    + exp.replace(ipReg, function(ss) {
                        var name = ss.split(/\$|\@/)[1];
                        if (/^\w+(\.\w+)*$/.test(name)) {
                            return ' serverVariables["' + name.toUpperCase() + '"]';
                        } else {
                            return ' parameters["' + name.toUpperCase() + '"]';
                        }
                    })
                    + '})()';
                p = iReg.test(s) ? 'escapeId(' + p + ').replace(/`/g, "")' : 'escape(' + p + ')';

                return '"\n+' + p + '+\n"';
            })
            .replace(/\n\s*/g, ' ');

            var TAB = '  ';
            fragment.push('/********');
            fragment.push(' * SQL ' + index);
            fragment.push(' ********/');
            fragment.push('transaction.push(function(callback) {');
            fragment.push('  var sql ');
            fragment.push('  try {')
            fragment.push('    sql = ' + sql);
            fragment.push('  } catch (ex) {');
            fragment.push('    ex.message = "Invalid sql params expression. Caused by: " + ex.message');
            fragment.push('    ex.debugInfo = {sqlIndex: ' + index + ', sql: sql, param: parameters}');
            fragment.push('    return callback(ex)');
            fragment.push('  }');
            fragment.push('  logger.debug("SQL ' + index + ': " , "\\n", sql)');
            fragment.push('  logger.debug("PARAMS: " , "\\n", JSON.stringify(parameters))');
            fragment.push('  databaseConnection.execute(sql, [], function(err, result) {');
            fragment.push('    if (err) {\n');
            fragment.push('      err.debugInfo = {sqlIndex: ' + index + ', sql: sql, params: parameters, originErrorMessage: err.message}');
            fragment.push('      err.message += " SQL index: ' + index + '"');
            fragment.push('      return callback(err)');
            fragment.push('    }');
            fragment.push('    $' + index + ' = result');
            fragment.push('    callback(null, result)');
            fragment.push('  })');
            fragment.push('})');
            return TAB + fragment.join('\n' + TAB);
        });
        var template = sideEffectsSQLCount > 1
            ? procedureTemplateTransaction
            : procedureTemplate;

        var funcStr = template
            .replace('{{TRANSACTION-BLOCKS}}', fragments.join('\n\n'))
            .replace('{{SERIES}}', series ? 'series' : 'parallel');

        return eval( funcStr );
    }
};

module.exports = compiler;